#include <algorithm> // std::sort
#include <numeric>   // std::accumulate

#include "cdo_remap.h"


const auto is_not_equal = [](auto a, auto b) noexcept { return  (a < b || b < a); };
const auto is_equal = [](auto a, auto b) noexcept { return !(a < b || b < a); };

extern "C"
{
#include "clipping.h"
#include "area.h"
#include "geometry.h"
}

struct CellSearch
{
  enum yac_edge_type *edgeType = nullptr;
  size_t numCellCorners = 0;
  size_t maxCells = 0;
  Varray<double> partialAreas;
  Varray<struct grid_cell> gridCells;
  Varray<struct grid_cell> overlapCells;
};

static void
cellsearch_realloc(size_t numCells, CellSearch &search)
{
  if (numCells > search.maxCells)
    {
      search.partialAreas.resize(numCells);
      search.overlapCells.resize(numCells);
      search.gridCells.resize(numCells);

      for (size_t i = search.maxCells; i < numCells; ++i)
        {
          search.overlapCells[i].array_size = 0;
          search.overlapCells[i].num_corners = 0;
          search.overlapCells[i].edge_type = nullptr;
          search.overlapCells[i].coordinates_xyz = nullptr;

          search.gridCells[i].array_size = search.numCellCorners;
          search.gridCells[i].num_corners = search.numCellCorners;
          search.gridCells[i].edge_type = search.edgeType;
          search.gridCells[i].coordinates_xyz = new double[search.numCellCorners][3];
        }

      search.maxCells = numCells;
    }
}

static void
cellsearch_free(CellSearch &search)
{
  for (size_t i = 0; i < search.maxCells; i++)
    {
      if (search.overlapCells[i].array_size > 0)
        {
          if (search.overlapCells[i].coordinates_xyz) free(search.overlapCells[i].coordinates_xyz);
          if (search.overlapCells[i].edge_type) free(search.overlapCells[i].edge_type);
        }

      delete[] search.gridCells[i].coordinates_xyz;
    }
}

static inline double
gridcell_area(const grid_cell &cell)
{
  return yac_huiliers_area(cell);
}

static void
cdo_compute_overlap_areas(size_t numSearchCells, CellSearch &search, const GridCell &gridCell)
{
  auto &overlapCells = search.overlapCells;

  // Do the clipping and get the cell for the overlapping area
  yac_cell_clipping(numSearchCells, search.gridCells.data(), gridCell.yacGridCell, overlapCells.data());

  // Get the partial areas for the overlapping regions
  for (size_t i = 0; i < numSearchCells; i++) search.partialAreas[i] = gridcell_area(overlapCells[i]);

#ifdef VERBOSE
  for (size_t i = 0; i < numSearchCells; i++) printf("overlap area : %lf\n", search.partialAreas[i]);
#endif
}

static double
get_edge_direction(double * ref_corner, double * corner_a, double * corner_b)
{
  double edge_norm[3];
  crossproduct_ld(corner_a, corner_b, edge_norm);
  normalise_vector(edge_norm);

  // sine of the angle between the edge and the reference corner
  double angle =
    edge_norm[0] * ref_corner[0] +
    edge_norm[1] * ref_corner[1] +
    edge_norm[2] * ref_corner[2];

  // if the reference corner is directly on the edge
  // (for small angles sin(x)==x)
  if (fabs(angle) < yac_angle_tol) return 0.0;

  return copysign(1.0, angle);
}

static void
cdo_compute_concave_overlap_areas(size_t numSearchCells, CellSearch &search, const GridCell &gridCell)
{
  const auto targetCell = gridCell.yacGridCell;
  auto &overlapAreas = search.partialAreas;
  auto &overlapCells = search.overlapCells;
  auto &sourceCells = search.gridCells;

  // common node point to all partial target cells
  double *baseCorner = targetCell.coordinates_xyz[0];

  // the triangulation algorithm only works for cells that only have great circle edges
  enum yac_edge_type edgeTypes[3] = { GREAT_CIRCLE_EDGE, GREAT_CIRCLE_EDGE, GREAT_CIRCLE_EDGE };
  double coordinates_xyz[3][3] = { { -1.0, -1.0, -1.0 }, { -1.0, -1.0, -1.0 }, { -1.0, -1.0, -1.0 } };
  coordinates_xyz[0][0] = baseCorner[0];
  coordinates_xyz[0][1] = baseCorner[1];
  coordinates_xyz[0][2] = baseCorner[2];

  // data structure to hold the triangles of the target cell
  struct grid_cell partialCell;
  partialCell.array_size = 3;
  partialCell.num_corners = 3;
  partialCell.coordinates_xyz = coordinates_xyz;
  partialCell.edge_type = edgeTypes;

  // Do the clipping and get the cell for the overlapping area

  for (size_t i = 0; i < numSearchCells; ++i) overlapAreas[i] = 0.0;

  // for all triangles of the target cell
  // (triangles a formed by first corner of the target cells and each edge of
  //  the cell; the first and last edge of the cell already, contain the
  //  first corner, therefore we can skip them)
  for (size_t cornerIdx = 1; cornerIdx < targetCell.num_corners - 1; ++cornerIdx)
    {
      const auto cornerA = targetCell.coordinates_xyz[cornerIdx];
      const auto cornerB = targetCell.coordinates_xyz[(cornerIdx + 1)];

      // if the current edge has a length of zero
      if (points_are_identically(cornerA, cornerB)) continue;

      const auto edgeDirection = get_edge_direction(baseCorner, cornerA, cornerB);

      partialCell.coordinates_xyz[1][0] = cornerA[0];
      partialCell.coordinates_xyz[1][1] = cornerA[1];
      partialCell.coordinates_xyz[1][2] = cornerA[2];
      partialCell.coordinates_xyz[2][0] = cornerB[0];
      partialCell.coordinates_xyz[2][1] = cornerB[1];
      partialCell.coordinates_xyz[2][2] = cornerB[2];

      // clip the current target cell triangle with all source cells
      yac_cell_clipping(numSearchCells, sourceCells.data(), partialCell, overlapCells.data());

      // Get the partial areas for the overlapping regions as sum over the partial target cells.
      for (size_t i = 0; i < numSearchCells; ++i)
        {
          if (overlapCells[i].num_corners == 0) continue;

          overlapAreas[i] += gridcell_area(overlapCells[i]) * edgeDirection;
        }
    }

  for (size_t i = 0; i < numSearchCells; ++i)
    {
      if (overlapAreas[i] < 0.0) overlapAreas[i] = -overlapAreas[i];
    }
 
#ifdef VERBOSE
  for (size_t i = 0; i < numSearchCells; i++) printf("overlap area %zu: %lf\n", i, overlapAreas[i]);
#endif
}

static void
set_cell_coordinates_yac(const size_t cellIndex, const size_t numCorners, const GridInfo &gridInfo,
                         const struct grid_cell &yacGridCell, double *x, double *y)
{
  const auto storeXY = (x && y);
  auto xyz = yacGridCell.coordinates_xyz;

  const auto cellCornersLon = &gridInfo.cellCornersLon[cellIndex * numCorners];
  const auto cellCornersLat = &gridInfo.cellCornersLat[cellIndex * numCorners];
  for (size_t i = 0; i < numCorners; ++i) LL_to_XYZ(cellCornersLon[i], cellCornersLat[i], xyz[i]);

  if (storeXY)
    {
      for (size_t k = 0; k < numCorners; ++k)
        {
          x[k] = cellCornersLon[k];
          y[k] = cellCornersLat[k];
        }
    }
}

static void
set_cell_coordinates(const size_t cellIndex, const size_t numCorners, const GridInfo &gridInfo, const GridCell &gridCell)
{
  set_cell_coordinates_yac(cellIndex, numCorners, gridInfo, gridCell.yacGridCell, gridCell.coordinates_x, gridCell.coordinates_y);
}

static void
set_coordinates_yac(const size_t numCells, const Varray<size_t> &cellIndices,
                    const size_t numCorners, const GridInfo &gridInfo, const Varray<grid_cell> &yacGridCell)
{
  for (size_t i = 0; i < numCells; ++i)
    set_cell_coordinates_yac(cellIndices[i], numCorners, gridInfo, yacGridCell[i], nullptr, nullptr);
}

static void
set_coordinates_yac(const double (*xyzCoords)[3], const size_t numCells, const Varray<size_t> &cellIndices,
                    const size_t numCorners, const Varray<grid_cell> &yacGridCell)
{
  for (size_t i = 0; i < numCells; ++i)
    {
      const auto offset = cellIndices[i] * numCorners;
      auto xyz = yacGridCell[i].coordinates_xyz;
      for (size_t k = 0; k < numCorners; ++k)
        for (size_t l = 0; l < 3; ++l)
          xyz[k][l] = xyzCoords[offset + k][l];
    }
}

static void
gridcell_init_yac(GridCell &gridCell, size_t numCorners, enum yac_edge_type *edgeType)
{
  gridCell.yacGridCell.array_size = numCorners;
  gridCell.yacGridCell.num_corners = numCorners;
  gridCell.yacGridCell.edge_type = edgeType;
  gridCell.yacGridCell.coordinates_xyz = new double[numCorners][3];
  gridCell.coordinates_x = new double[numCorners];
  gridCell.coordinates_y = new double[numCorners];
}

static void
gridcell_free_yac(const GridCell &gridCell)
{
  delete[] gridCell.yacGridCell.coordinates_xyz;
  delete[] gridCell.coordinates_x;
  delete[] gridCell.coordinates_y;
}

static size_t
remove_invalid_areas(size_t numSearchCells, Varray<double> &areas, Varray<size_t> &searchIndices)
{
  size_t n = 0;
  for (size_t i = 0; i < numSearchCells; ++i)
    {
      if (areas[i] > 0.0)
        {
          areas[n] = areas[i];
          searchIndices[n] = searchIndices[i];
          n++;
        }
    }

  return n;
}

static size_t
remove_invalid_weights(size_t gridSize, size_t numWeights, Varray<double> &weights, Varray<size_t> &searchIndices)
{
  size_t n = 0;
  for (size_t i = 0; i < numWeights; ++i)
    {
      const auto cellIndex = (weights[i] > 0.0) ? searchIndices[i] : gridSize;
      if (cellIndex != gridSize)
        {
          weights[n] = weights[i];
          searchIndices[n] = cellIndex;
          n++;
        }
    }

  return n;
}

static size_t
remove_unmask_weights(const Varray<short> &gridMask, size_t numWeights, Varray<double> &weights,
                      Varray<size_t> &searchIndices)
{
  size_t n = 0;
  for (size_t i = 0; i < numWeights; ++i)
    {
      const auto cellIndex = searchIndices[i];
      /*
        Store the appropriate addresses and weights.
        Also add contributions to cell areas.
        The source grid mask is the master mask.
      */
      if (gridMask[cellIndex])
        {
          weights[n] = weights[i];
          searchIndices[n] = cellIndex;
          n++;
        }
    }

  return n;
}

static void
normalize_weights(NormOpt normOpt, double cellArea, double cellFrac, size_t numWeights, Varray<double> &weights)
{
  if (normOpt == NormOpt::DESTAREA)
    {
      const auto normFactor = is_not_equal(cellArea, 0.0) ? 1.0 / cellArea : 0.0;
      for (size_t i = 0; i < numWeights; ++i) weights[i] *= normFactor;
    }
  else if (normOpt == NormOpt::FRACAREA)
    {
      const auto normFactor = is_not_equal(cellFrac, 0.0) ? 1.0 / cellFrac : 0.0;
      for (size_t i = 0; i < numWeights; ++i) weights[i] *= normFactor;
    }
  else if (normOpt == NormOpt::NONE)
    {
    }
}

static void
correct_weights(double cellArea, size_t numWeights, Varray<double> &weights)
{
  for (size_t i = 0; i < numWeights; ++i) weights[i] /= cellArea;
  yac_correct_weights(numWeights, weights.data());
  for (size_t i = 0; i < numWeights; ++i) weights[i] *= cellArea;
}

static void
sort_weights(size_t numWeights, Varray<size_t> &searchIndices, Varray<double> &weights)
{
  size_t n;
  for (n = 1; n < numWeights; ++n)
    if (searchIndices[n] < searchIndices[n - 1]) break;
  if (n == numWeights) return;

  if (numWeights > 1)
    {
      struct IndexWeightX
      {
        size_t index;
        double weight;
      };

      Varray<IndexWeightX> indexWeights(numWeights);

      for (n = 0; n < numWeights; ++n)
        {
          indexWeights[n].index = searchIndices[n];
          indexWeights[n].weight = weights[n];
        }

      const auto compareIndex = [](const auto &a, const auto &b) noexcept { return a.index < b.index; };
      std::sort(indexWeights.begin(), indexWeights.end(), compareIndex);

      for (n = 0; n < numWeights; ++n)
        {
          searchIndices[n] = indexWeights[n].index;
          weights[n] = indexWeights[n].weight;
        }
    }
}

static int
get_lonlat_circle_index(size_t numCells, size_t numCorners, const Varray<double> &clon, const Varray<double> &clat)
{
  int lonlatCircleIndex = -1;

  if (numCorners == 4)
    {
      size_t iadd = (numCells < 100) ? 1 : numCells / 30 - 1;
      size_t num_i = 0, num_eq0 = 0, num_eq1 = 0;

      for (size_t i = 0; i < numCells; i += iadd)
        {
          const auto i4 = i * 4;
          num_i++;
          // clang-format off
          if (is_equal(clon[i4 + 1], clon[i4 + 2]) && is_equal(clon[i4 + 3], clon[i4 + 0]) &&
              is_equal(clat[i4 + 0], clat[i4 + 1]) && is_equal(clat[i4 + 2], clat[i4 + 3]))
            {
              num_eq1++;
            }
          else if (is_equal(clon[i4 + 0], clon[i4 + 1]) && is_equal(clon[i4 + 2], clon[i4 + 3]) &&
                   is_equal(clat[i4 + 1], clat[i4 + 2]) && is_equal(clat[i4 + 3], clat[i4 + 0]))
            {
              num_eq0++;
            }
          // clang-format on
        }

      if (num_i == num_eq1) lonlatCircleIndex = 1;
      if (num_i == num_eq0) lonlatCircleIndex = 0;
    }

  // printf("lonlatCircleIndex %d\n", lonlatCircleIndex);

  return lonlatCircleIndex;
}

static int
get_lonlat_circle_index(const GridInfo &gridInfo)
{
  return get_lonlat_circle_index(gridInfo.numCells, gridInfo.numCorners, gridInfo.cellCornersLon, gridInfo.cellCornersLat);
}

static double
conserv_remap(const Varray<double> &srcArray, size_t numWeights, const Varray<double> &weights, const Varray<size_t> &srcIndices)
{
  double tgtPoint = 0.0;
  for (size_t i = 0; i < numWeights; ++i) tgtPoint += srcArray[srcIndices[i]] * weights[i];

  return tgtPoint;
}

void
remap_conserv(NormOpt normOpt, const RemapSearch &remapSearch, const Varray<double> &srcArray, Varray<double> &tgtArray, double missval)
{
  auto &srcGrid = remapSearch.srcGrid;
  auto &tgtGrid = remapSearch.tgtGrid;

  auto srcNumCells = srcGrid.numCells;
  auto tgtNumCells = tgtGrid.numCells;

  Varray<short> srcGridMask(srcNumCells);
  for (size_t i = 0; i < srcNumCells; ++i) srcGridMask[i] = !is_equal(srcArray[i], missval);

  auto srcNumCorners = srcGrid.numCorners;
  auto tgtNumCorners = tgtGrid.numCorners;

  enum yac_edge_type lonlatCircleType[] = { LON_CIRCLE_EDGE, LAT_CIRCLE_EDGE, LON_CIRCLE_EDGE, LAT_CIRCLE_EDGE, LON_CIRCLE_EDGE };
  Varray<enum yac_edge_type> greatCircleType(std::max(srcNumCorners, tgtNumCorners), GREAT_CIRCLE_EDGE);

  auto srcEdgeType = greatCircleType.data();
  auto tgtEdgeType = greatCircleType.data();

  enum yac_cell_type tgtCellType = MIXED_CELL;

  if (srcNumCorners == 4)
    {
      const auto lonlatCircleIndex = get_lonlat_circle_index(srcGrid);
      if (lonlatCircleIndex >= 0) srcEdgeType = &lonlatCircleType[lonlatCircleIndex];
    }

  if (tgtNumCorners == 4)
    {
      const auto lonlatCircleIndex = get_lonlat_circle_index(tgtGrid);
      if (lonlatCircleIndex >= 0)
        {
          tgtCellType = LON_LAT_CELL;
          tgtEdgeType = &lonlatCircleType[lonlatCircleIndex];
        }
    }

  GridCell tgtGridCell;
  gridcell_init_yac(tgtGridCell, tgtNumCorners, tgtEdgeType);

  CellSearch cellSearch;
  cellSearch.numCellCorners = srcNumCorners;
  cellSearch.edgeType = srcEdgeType;

  auto numCellCorners = srcNumCorners;  // num of corners of search cells

  Varray<size_t> searchIndices(srcNumCells);

  // Loop over target grid
  for (size_t tgtCellIndex = 0; tgtCellIndex < tgtNumCells; ++tgtCellIndex)
    {
      tgtArray[tgtCellIndex] = missval;

      //if (!tgtGrid.mask[tgtCellIndex]) continue;

      set_cell_coordinates(tgtCellIndex, tgtNumCorners, tgtGrid, tgtGridCell);

      // Get search cells
      auto numSearchCells = do_gridcell_search(remapSearch.gcs, false, tgtGridCell, searchIndices);

      if (numSearchCells == 0) continue;

      // Create search arrays

      cellsearch_realloc(numSearchCells, cellSearch);

      if (remapSearch.gcs.xyzCoords)
        set_coordinates_yac(remapSearch.gcs.xyzCoords, numSearchCells, searchIndices, numCellCorners, cellSearch.gridCells);
      else
        set_coordinates_yac(numSearchCells, searchIndices, numCellCorners, srcGrid, cellSearch.gridCells);

      if (tgtNumCorners < 4 || tgtCellType == LON_LAT_CELL)
        cdo_compute_overlap_areas(numSearchCells, cellSearch, tgtGridCell);
      else
        cdo_compute_concave_overlap_areas(numSearchCells, cellSearch, tgtGridCell);

      auto &partialWeights = cellSearch.partialAreas;

      auto numWeights = remove_invalid_areas(numSearchCells, partialWeights, searchIndices);

      const auto tgtCellArea = gridcell_area(tgtGridCell.yacGridCell);

      if (normOpt == NormOpt::FRACAREA) correct_weights(tgtCellArea, numWeights, partialWeights);

      numWeights = remove_invalid_weights(srcNumCells, numWeights, partialWeights, searchIndices);
      numWeights = remove_unmask_weights(srcGridMask, numWeights, partialWeights, searchIndices);

      const auto tgtCellFrac = std::accumulate(partialWeights.begin(), partialWeights.begin() + numWeights, 0.0);

      if (numWeights)
        {
          sort_weights(numWeights, searchIndices, partialWeights);
          // Normalize weights using target cell area if requested
          normalize_weights(normOpt, tgtCellArea, tgtCellFrac, numWeights, partialWeights);
          tgtArray[tgtCellIndex] = conserv_remap(srcArray, numWeights, partialWeights, searchIndices);
        }
    }

  // Finished with all cells: deallocate search arrays

  cellsearch_free(cellSearch);
  gridcell_free_yac(tgtGridCell);
}  // remap_conserv
